import java.io.IOException;
import java.io.BufferedReader;
import java.nio.charset.Charset;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Paths;
import java.nio.file.StandardCopyOption;
import java.nio.file.Path;
import java.nio.file.DirectoryStream;
import java.nio.file.FileVisitOption;
import java.nio.file.SimpleFileVisitor;
import java.nio.file.FileVisitResult;
import java.nio.file.attribute.BasicFileAttributes;
import java.nio.file.attribute.BasicFileAttributeView;
import java.nio.file.attribute.FileTime;
import java.nio.charset.StandardCharsets;
import java.util.concurrent.TimeUnit;
import java.util.stream.Stream;

public class DemoPath {

  public static void main(String[] args) throws IOException {
    //demoPath(); // Path interfece which is new in Java 7
    //demoReadFileJ11(); // using Java 11 features
    //demoFiles(); // using Java 7+ Files and Paths utility classes
    //demoTraversal(); // traverse (visit) directories
    //demoBufferedOps(); // convenient reading/writing text to file
    demoStreams(); // Java 8 Streams demo
  }

  /** Path - An object that may be used to locate a file in a file system. */
  public static void demoPath() throws IOException {
    // identify folder/subfolder
    Path subfldr = Path.of("folder", "subfolder");

    // identify file in the subfolder
    Path file = subfldr.resolve("file.txt");

    // simply identify a sibling file (ie. in same directory/folder)
    Path siblingFile = file.resolveSibling("secondfile.txt");

    System.out.println("Subfolder path resolves to: " + subfldr);
    System.out.println("File path resolves to: " + file);
    System.out.println("Sibling file resolves to: " + siblingFile);

    // Iterate over path segments
    System.out.println("Sibling file path segments:");
    for (Path segment : siblingFile) {
      System.out.print(" segment=" + segment);
    }

    // crazy paths can also be resolved
    Path relative = Path.of("..", "subfolder");
    Path subfldrUgly = subfldr.resolve(relative);
    System.out.println("\nSubfolder path ugly: " + subfldrUgly);
    // normalize it
    Path subfldrNorm = subfldrUgly.normalize();
    System.out.println("Subfolder path normalized: " + subfldrNorm);
  }

  public static void demoReadFileAsString() throws IOException {
    Path path = Paths.get("text.txt");
    String s = Files.readString(path);
    System.out.println("Content read from text.txt: " + s);
    
    s = Files.readString(path, StandardCharsets.UTF_8);
    System.out.println("Content read from text.txt: " + s);
  }

  public static void demoFiles() throws IOException {
    Path src = Path.of("folder", "subfolder", "file.txt");
    Path dest = Path.of("folder", "subfolder", "dest.txt");

    System.out.println("Deleted dest.txt before copy? " + Files.deleteIfExists(dest));

    Files.copy(src, dest, StandardCopyOption.COPY_ATTRIBUTES);
    // Files.move works per analogy

    BasicFileAttributes a = Files.readAttributes(dest, BasicFileAttributes.class);
    System.out.println("Attributes of dest.txt: " + a);
    FileTime ftcr = a.creationTime(); // Returns the creation time.
    System.out.println("creationTime: " + ftcr);

    Object o = a.fileKey(); // Returns an object that uniquely identifies the given file, or null if a file
                            // key is not available.
    System.out.println("fileKey: " + o);

    boolean dir = a.isDirectory(); // Tells whether the file is a directory.
    System.out.println("isDirectory? " + dir);

    boolean oth = a.isOther(); // Tells whether the file is something other than a regular file, directory, or
                               // symbolic link.
    System.out.println("isOther? " + oth);

    boolean regf = a.isRegularFile(); // Tells whether the file is a regular file with opaque content.
    System.out.println("isRegularFile? " + regf);

    boolean sl = a.isSymbolicLink(); // Tells whether the file is a symbolic link.
    System.out.println("isSymbolicLink? " + sl);

    FileTime ftla = a.lastAccessTime(); // Returns the time of last access.
    System.out.println("lastAccessTime: " + ftla);

    FileTime ftlm = a.lastModifiedTime(); // Returns the time of last modification.
    System.out.println("lastModifiedTime: " + ftlm);
    System.out.println("lastModifiedTime.to(HOURS): " + ftlm.to(TimeUnit.HOURS));
    System.out.println("lastModifiedTime.toMillis(): " + ftlm.toMillis());

    long size = a.size(); // Returns the size of the file (in bytes).
    System.out.println("size: " + size);
  }

  public static void demoTraversal() throws IOException {
    Path dir = Path.of(".");
    try (DirectoryStream<Path> stream = Files.newDirectoryStream(dir, "*.{java,class}")) {
      for (Path entry : stream) {
        System.out.println(entry);
      }
    } 

    System.out.println("Recursive walk through files in . (max depth 2)");
    try (Stream<Path> stream = Files.walk(dir, /* max depth */ 2)) {
      stream.forEach(System.out::println);
    } 

    System.out.println("Recursive walk using FileVisitor");
    Files.walkFileTree​(dir, new SimpleFileVisitor<Path>() {
      @Override
      public FileVisitResult preVisitDirectory(Path dir, BasicFileAttributes attrs) throws IOException {
        System.out.println("preVisitDirectory " + dir);
        return FileVisitResult.CONTINUE;
      }

      @Override
      public FileVisitResult visitFile(Path file, BasicFileAttributes attrs) throws IOException {
        System.out.println("visitFile " + file);
        return FileVisitResult.CONTINUE;
      }

      @Override
      public FileVisitResult visitFileFailed(Path file, IOException exc) throws IOException {
        System.out.println("visitFileFailed " + file);
        return FileVisitResult.CONTINUE;
      }

      @Override
      public FileVisitResult postVisitDirectory(Path dir, IOException exc) throws IOException {
        System.out.println("postVisitDirectory " + dir);
        return FileVisitResult.CONTINUE;
      }
    });
  }

  public static void demoBufferedOps() throws IOException {
    try (BufferedReader br = Files.newBufferedReader(Path.of("text.txt"))) {
      String s = br.readLine();
      int i = 1;
      System.out.println("Reading file ./text.txt");
      while (s != null) {
        System.out.println("Line " + i + ". " + s);
        s = br.readLine();
        i++;
      }
    }
  }

  public static void demoStreams() throws IOException {
    try (Stream<String> stream = Files.lines(Path.of("text.txt"))) {
      // sumarizes 3rd fields (as integers)
      int sum = stream
          .filter(x -> x.startsWith("H")) // take only those H* greetings
          .mapToInt(x -> Integer.parseInt(x.split(" ")[2])) // take third field (number)
          .sum(); // sum the numbers
      System.out.println("Sum of greetings H* " + sum);
    }
  }
}
